/*
 * CO2mini control program
 * 2017-01-07 Kouichi Hirabayashi
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <err.h>
#include <ctype.h>
#include <errno.h>
#include <libusb.h>
#include <usbhid.h>

#define MANUFACTURER_ID 0x4d9		// Holtek
#define PRODUCT_ID	0xa052		// USB-zyTemp

typedef unsigned char co2mon_data_t[8];

int hid_open(unsigned short, unsigned short, int);
void swap_char(unsigned char *, unsigned char *);
void decode_buf(co2mon_data_t, co2mon_data_t, co2mon_data_t);
void error(char *fmt, ...);
void Usage();
char *strsave(char *);

co2mon_data_t magic_table = { 0 };
co2mon_data_t data = { 0 };
co2mon_data_t val;

int N;		// using Nth CO2mini device
char *cmd;

int main(int argc, char **argv)
{
  char *s;
  char dev[64];
  int fd, i, id, k, n, cksum, rsiz, w;
  double Tamb = -1e3;
  int CntR = -100;
  report_desc_t r;
  hid_data_t hd;
  hid_item_t shi, rhi;
  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-' && !isdigit((*argv)[1]))
	for (s = argv[0]+1; *s; s++)
		switch (*s) {
		defualt:
			Usage();
		}
  if (argc > 0){
	sscanf(argv[0], "%d", &N);
	if (--N < 0)
		Usage();
  }

  hid_init(NULL);

  if ((fd = hid_open((unsigned short)MANUFACTURER_ID,
		(unsigned short)PRODUCT_ID, N)) < 0)
	error("can't open /dev/hid%d.\n", N + 1);

  if ((id = hid_get_report_id(fd)) < 0) 
	perror("hid_get_report_id():");
  if ((r = hid_get_report_desc(fd)) < 0)
	errx(1, "USB_GET_REPORT_DESC");
  rsiz = hid_report_size(r, 1<<hid_output, id);
  hd = hid_start_parse(r, 1<<hid_output, id);
  hid_get_item(hd, &shi);
  do {
	if(shi.kind == hid_output)
		break;
  } while (hid_get_item(hd, &shi) > 0);
  hid_end_parse(hd);
  hid_set_report(fd, 1<<hid_output, (unsigned char *)magic_table, 8);
  write(fd, magic_table, sizeof(co2mon_data_t));

  k = 0;	// error counter
  do {
	for (hd = hid_start_parse(r, 1<<hid_input, id);
			hid_get_item(hd, &rhi); ) {
		if(rhi.kind == hid_input)
			break;
	}
	hid_end_parse(hd);
	/* clear file inputs */
	n = read(fd, data, BUFSIZ);
	decode_buf(val, data, magic_table);
	if (val[4] != 0x0D) {
		if (++k > 3)
			fprintf(stderr, "Unexpected data (%02X) != 0x0D.\n",
				val[4]);
		continue;
	}
	cksum = (val[0] + val[1] + val[2]) & 0xFF;
	if (cksum != val[3]) {
		if (++k > 3)
			fprintf(stderr, "checksum error (%02X) != %02X\n",
				cksum, val[3]);
		continue;
	}
	w = (val[1] <<8) + val[2];
	switch (val[0]) {
	case 'B':	// 0x42
		Tamb = (double)w / 16.0 - 273.15;
		break;
	case 'P':	// 0x50
		CntR = (int)w;
		break;
#if 0
	case 'D':	// 0x44
		RH = (int)(w * 1e2);
		printf("RH %4.0f\n", (int)(w * 1e2));
		break;
#endif
	default:
		// ignore
		break;
	}
  } while (Tamb == -1e3 || CntR == -100);
  printf("%.1f C %d ppm\n", Tamb, CntR);
  hid_dispose_report_desc(r);
  close(fd);

  exit(0);
}

void swap_char(unsigned char *a, unsigned char *b)
{
  unsigned char tmp = *a;

  *a = *b;
  *b = tmp;
}

void decode_buf(co2mon_data_t val, co2mon_data_t buf, co2mon_data_t magic_table)
{
  const unsigned char magic_word[8] = "Htemp99e";
  int i;
  unsigned char tmp;

  swap_char(&buf[0], &buf[2]);
  swap_char(&buf[1], &buf[4]);
  swap_char(&buf[3], &buf[7]);
  swap_char(&buf[5], &buf[6]);

  tmp = (buf[7] << 5);
  val[7] = (buf[6] << 5) | (buf[7] >> 3);
  val[6] = (buf[5] << 5) | (buf[6] >> 3);
  val[5] = (buf[4] << 5) | (buf[5] >> 3);
  val[4] = (buf[3] << 5) | (buf[4] >> 3);
  val[3] = (buf[2] << 5) | (buf[3] >> 3);
  val[2] = (buf[1] << 5) | (buf[2] >> 3);
  val[1] = (buf[0] << 5) | (buf[1] >> 3);
  val[0] = tmp | (buf[0] >> 3);
  for (i = 0; i < 8; ++i) {
	val[i] -= (magic_word[i] << 4) | (magic_word[i] >> 4);
  }
}

dump(s, n) char *s; int n;
{
  int i;

  for (i = 0; i < n; i++) {
	printf(" %02x", s[i] & 0xff);
  }
  printf("\n");
  return 0;
}

void error(char *fmt, ...)
{
  va_list argp;

  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(1);
}

void Usage()
{
  static char *t[] = {
"Commad %s - CO2mini access program",
"Usage	%s [n]",
"Example	%s	# print ambient temperature(C) and CO2 concentration (ppm)",
"		%s 2	# using second CO2mini device",
  NULL
};
  int i;

  for (i = 0; t[i] != NULL; i++) {
	fprintf(stderr, t[i], cmd);
	fprintf(stderr, "\n");
  }
  exit(1);
}	

int hid_open(unsigned short idVendor, unsigned short idProduct, int n)
{
  int fd, i, id, rsiz;
  report_desc_t r;
  hid_data_t hd;
  hid_item_t shi, rhi;
  char dev[16];		// device file name
  int nDev = 0;
#define MAX_DEV		50
  char *Dev[MAX_DEV];	// CO2mini device

  for (i = 0; i < MAX_DEV; i++) {
	sprintf(dev, "/dev/uhid%d", i);
	if ((fd = open(dev, O_RDWR)) < 0)
		break;
	if ((id = hid_get_report_id(fd)) < 0) 
                perror("hid_get_report_id():");
        r = hid_get_report_desc(fd);
        if (r == 0) 
                errx(1, "USB_GET_REPORT_DESC");
	rsiz = hid_report_size(r, 1<<hid_output, id);
	hd = hid_start_parse(r, 1<<hid_output, id);
	hid_get_item(hd, &shi);
	if (shi._usage_page == 0xff000000 &&
			shi.usage == 0xff000001 && rsiz == 8)
		Dev[nDev++] = strsave(dev);
	hid_end_parse(hd);
	close(fd);
  }
  if (i == MAX_DEV)
	fd = -1;
//  for (i = 0; i < nDev; i++)
//	printf("%d %s\n", i, Dev[i]);
  fd = open(Dev[n], O_RDWR);
  return fd;
}

char *strsave(char *s)
{
  char *p;

  if ((p = (char *)malloc(strlen(s) + 1)) == NULL) {
	fprintf(stderr, "strsave(): mamory over.\n");
	exit(1);
  }
  return strcpy(p, s);
}
